#include <iostream>
#include <string>
#include <vector>

//
// La classe TrOp contiene una operazione che puo'
// essere annullata durante un rollback.
// Si tratta di una classe base astratta da cui
// derivano tutti i tipi di operazione annullabile.
//
class TrOp
{
  public:
    virtual void Undo() = 0;
    virtual ~TrOp() {}
};

//
// Il template Change genera al variare del tipo
// dati usato le possibili operazioni (derivate
// da TrOp) che e' possibile annullare in un
// rollback. In pratica l'istruzione
//
//           new Change<int>(a);
//
// crea un oggetto "cambio di intero" e salva
// il vecchio valore in modo da rendere
// annullabile l'operazione eseguita.
//
template<class T> class Change : public TrOp
{
  public:
    T& who;
    T old_value;
    
    Change(T& s) : who(s), old_value(s) { }
    void Undo() { who=old_value; }
};

//
// Il vettore Transaction contiene tutte le
// operazioni da annullare in caso di una
// richiesta di rollback.
//
std::vector<TrOp *> Transaction;

//
// Il template field genera al variare del
// tipo dati i possibili campi di un oggetto.
// Il campo provvede a generare un opportuno
// oggetto Change<tipo> ad ogni modifica.
//
template<class T> class Field : public T
{
  private:
    void Save()
    {
      Transaction.push_back(new Change<T>(*static_cast<T*>(this)));
    }
    Field& operator&();
  
  public:
    // Costruttore da oggetto T
    Field(const T& s) : T(s) { }
    
    // Conversione a valore (const!!!)
    operator const T() const { *this; }
    
    // Assegnazione
    Field& operator=(const T& v)
    { Save(); T::operator=(v); return *this; }
    
    Field& operator=(const Field<T>& v)
    { Save(); T::operator=(v); return *this; }
    
    // Rimappatura ++/--
    const T operator++(int) { Save(); return T::operator++(0); }
    const T operator--(int) { Save(); return T::operator--(0); }
    Field& operator++() { Save(); T::operator++(); return *this; }
    Field& operator--() { Save(); T::operator--(); return *this; }
    
    // Rimappatura incrementi/decrementi
    Field& operator+=(const T& v) { Save(); T::operator+=(v); return *this; }
    Field& operator-=(const T& v) { Save(); T::operator-=(v); return *this; }
    Field& operator*=(const T& v) { Save(); T::operator*=(v); return *this; }
    Field& operator/=(const T& v) { Save(); T::operator/=(v); return *this; }
    Field& operator%=(const T& v) { Save(); T::operator%=(v); return *this; }

    //
    // Nota: se sono presenti metodi che modificano una
    // classe e' necessario inserire un opportuno override
    // in modo da effettuare una Save() completa o
    // Save() specifiche delle parti modificate prima di
    // invocare il metodo in questione.
    //
};

//
// Template per i tipi base. Necessario perche' non e'
// possibile derivare da int e simili.
//
template<class T> class BasicField
{
  protected:
    T v_;
    void Save()
    {
      Transaction.push_back(new Change<T>(v_));
    }
    BasicField& operator&();
  
  public:
    // Costruttore di default
    BasicField() : v_(T()) { }

    // Costruttore da oggetto T
    BasicField(const T& s) : v_(s) { }
    
    // Conversione a valore (const!!!)
    operator const T() const { return v_; }
    
    // Assegnazione
    BasicField& operator=(const T& v)
    { Save(); v_=v; return *this; }
    
    BasicField& operator=(const BasicField<T>& v)
    { Save(); v_=v; return *this; }
    
    // Rimappatura ++/--
    const T operator++(int) { Save(); return v_++; }
    const T operator--(int) { Save(); return v_--; }
    BasicField& operator++() { Save(); ++v_; return *this; }
    BasicField& operator--() { Save(); --v_; return *this; }
    
    // Rimappatura incrementi/decrementi
    BasicField& operator+=(const T& v) { Save(); v_+=v; return *this; }
    BasicField& operator-=(const T& v) { Save(); v_-=v; return *this; }
    BasicField& operator*=(const T& v) { Save(); v_*=v; return *this; }
    BasicField& operator/=(const T& v) { Save(); v_/=v; return *this; }
    BasicField& operator%=(const T& v) { Save(); v_%=v; return *this; }
    
    // Rimappatura output
    friend std::ostream& operator << (std::ostream& o,const BasicField& t)
    { return o << t.v_; }

};

//
// Template simile a Field<> per i puntatori. Necessario perche' per
// qualche strano motivo la specializzazione di Field<> non compila
// su Microsoft VisualC++.
//
template<class T> class Ptr : public BasicField<T*>
{
  public:
    typedef T *TPtr;
    Ptr(const TPtr& x) : BasicField<T*>(x) {  }
    
    Ptr& operator=(const TPtr& x)
    { BasicField<T*>::operator=(x); return *this; }
    
    Ptr& operator=(const BasicField<T*>& x)
    { BasicField<T*>::operator=(x); return *this; }
    
    // Rimappatura dereferencing
    T  operator*()  { return *v_; }
    T* operator->() { return v_; }
};

//
// Specializzazione completa per Field<int>. Definizioni
// simili sono necessarie per gli altri tipi base.
//
template<> class Field<int> : public BasicField<int>
{
  public:
    Field(const int& x) : BasicField<int>(x) {  }
    Field& operator=(const int& x)
    { BasicField<int>::operator=(x); return *this; }
    Field& operator=(const BasicField<int>& x)
    { BasicField<int>::operator=(x); return *this; }
};

void Rollback()
{
  while (!Transaction.empty())
  {
    Transaction.back()->Undo();
    delete Transaction.back();
    Transaction.pop_back();
  }
}

int main()
{
  Field<std::string> s("Valore iniziale");
  Field<int> i(0);
  int j[4]={6,5,0,2};
  Ptr<int> p=j;
  
  p++;
  s+=" (modificato)";
  i+=3;
  i++;

  int *pp=p;

  std::cout << " s = " << s << std::endl;
  std::cout << " i = " << i << std::endl;
  std::cout << " *p = " << *p << std::endl;
  std::cout << " *pp = " << *pp << std::endl;

  Rollback();
  
  std::cout << " s = " << s.c_str() << std::endl;
  std::cout << " i = " << i << std::endl;
  std::cout << " *p = " << *p << std::endl;
  std::cout << " *pp = " << *pp << std::endl;

  std::cout << "sizeof(Field<int>) = " << sizeof(Field<int>) << std::endl;
  return 0;
}
